local MAX_LEVEL = 90
local GEAR = 'Gear'
local AURA = 'Aura'

local tblOops = {}              -- Primary warnings: bad stats, missing auras
local tblOops2 = {}             -- Secondary warnings: missing gems, missing enchants
local tblOops3 = {}             -- Tertiary warnings: missing raid buffs
local tblShoppingListGems = {}
local tblShoppingListEnch = {}
local tooltip = ''

local PLAIN_LETTER = 8383   -- For use in generating simple tooltips
local LibWeagleTooltip = LibStub("LibWeagleTooltip-2.1")

local THROTTLE = 1
local inCombat = false

-- Local references
local sanityCheck_config
local tblStatNames          = tblStatNames
local tblWrongStats         = tblWrongStats
local tblVerbs              = tblVerbs
local tblRoles              = tblRoles
local tblVerifyAuras        = tblVerifyAuras
local tblRaidBuffs          = tblRaidBuffs
local tblBuffTypesByName    = tblBuffTypesByName
local tblVerifyRaidBuffs    = tblVerifyRaidBuffs
local tblSlots              = tblSlots
local tblSlots_Enchantable  = tblSlots_Enchantable

local _G = _G

------------------------------------------------------------------
--      iter_cycle
do
    local function next (self)
        local ret = self.values[self.cur]
        self.cur = self.cur + 1
        if self.cur>#self.values then
            self.cur = 1
        end
        return ret
    end
   
    function iter_cycle (...)
        local obj = {
            values = {...},
            cur = 1,
            next = next,
        }
      
        return obj
    end   
end

------------------------------------------------------------------

local function VerifyItemStats (itemLink)
    -- Check against the bad stats table, return false for a bad item.
    local class = _G.UnitClass('player')
    local spec = _G.GetSpecialization('player')
    local tblStats = _G.GetItemStats (itemLink)
    
    if tblWrongStats[class]==nil  or  spec==nil then 
        return true
    end
    
    if tblStats==nil then tblStats={} end -- Just to prevent a stupid login error, really
    for stat,statVal in _G.pairs(tblStats) do
        for i,wrongStat in _G.pairs(tblWrongStats[class][spec]) do
            if stat==wrongStat then
                return false
            end
        end
    end
    
    return true
end

------------------------------------------------------------------

local function VerifyItemType (itemLink)
    -- For high level toons, also check ilvl and quality. The ilvl threshold is pretty arbitrary.
    -- Return false for a bad item.
    if _G.UnitLevel('player') == MAX_LEVEL then
        local _,_,qual, ilvl = _G.GetItemInfo (itemLink)
        if qual==nil or ilvl==nil then      -- This seems to happen occasionally... must be a bug of some sort of while logging on...
            return true
        end
        if qual<=1 or ilvl<=300 then
            return false
        end
    end
    
    return true
end

------------------------------------------------------------------

local function ShoppingList_CountSockets (itemLink)
    if not itemLink then return end
    
    local _,_,_, ilvl,_,_,_,_,equipSlot = _G.GetItemInfo(itemLink)
    if ilvl==nil or ilvl < sanityCheck_config['GemMinLvl'] then return end

    
    local tmp
    local i = 1
    while true do
        tmp = LibWeagleTooltip:GetTooltipLine(itemLink, i)
        if not tmp then 
            break
        else 
            if tmp:match ('Socket') and not tmp:match('Bonus') then 
                tblShoppingListGems[tmp] = (tblShoppingListGems[tmp] or 0) + 1
            end            
            i = i+1
        end
    end
end

------------------------------------------------------------------

local function CountSockets_Total (itemLink)
    if itemLink==nil then return 0 end
    local tblStats = _G.GetItemStats (itemLink)
    local ret = 0

    if tblStats==nil then tblStats={} end -- Just to prevent a stupid login error, really
    for k,v in _G.pairs(tblStats) do
        if k:match('SOCKET') then 
            ret = ret+v
        end
    end
    
    return ret
end

------------------------------------------------------------------

local function CountSockets_Gemmed (itemLink)
    if itemLink==nil then return 0 end
    local itemId, enchantId, gem1, gem2, gem3, gem4 = itemLink:match("item:(%d+):(%d+):(%d+):(%d+):(%d+):(%d+)")
    
    local ret = 0
    for i,v in _G.ipairs({gem1,gem2,gem3,gem4}) do
        if v~='0' then
            ret = ret+1
        end
    end
    
    return ret
end

------------------------------------------------------------------

local function StringHasBadStat (str)
    if str==nil then 
        return false
    end
    
    local class = _G.UnitClass('player')
    local spec = _G.GetSpecialization('player')
    if not spec then return end
    for i,stat in _G.pairs(tblWrongStats[class][spec]) do
        if str:match (tblStatNames[stat]) then
            return true
        end
    end
    
    return false
end

------------------------------------------------------------------

local function VerifySocketAndEnchantStats (itemLink)
    -- Generate a tooltip of a plain letter (statless, cached item) with the same enchant and gems as this item
    local itemId, enchantId, gem1, gem2, gem3, gem4 = itemLink:match("item:(%d+):(%d+):(%d+):(%d+):(%d+):(%d+)")
    local tooltip = ("item:%i:%i:%i:%i:%i:%i"):format(PLAIN_LETTER, enchantId, gem1, gem2, gem3, gem4)

    -- Get the enchant from line 2 of the tooltip, and up to 4 gems from the next 4 lines
    local tmp = {}  -- This table will contain the enchant text, followed by up to 4 gem texts
    for i=2,6 do
        tmp[i-1] =  LibWeagleTooltip:GetTooltipLine(tooltip, i)
    end
    
    for k,v in _G.pairs(tmp) do
        if StringHasBadStat(v) then
            return false
        end
    end

    return true
end

------------------------------------------------------------------

local function VerifyItemSockets (itemLink)
    local _,_,_, ilvl,_,_,_,_,equipSlot = _G.GetItemInfo(itemLink)
    
    if ilvl==nil or ilvl < sanityCheck_config['GemMinLvl'] then
        return true
    else
        return (CountSockets_Total(itemLink) <= CountSockets_Gemmed(itemLink))
    end
end

------------------------------------------------------------------

local function VerifyExtraSocket (slot)
    -- Only if the item is gemmed to begin with
    local itemLink = _G.GetInventoryItemLink ('player', slot)
    if not itemLink then return true end
    
    local _,_,_, ilvl = _G.GetItemInfo(itemLink)
    
    if ilvl==nil or ilvl < sanityCheck_config['EnchMinLvl'] then
        return true
    end
    
    return (CountSockets_Total(itemLink)+1 == CountSockets_Gemmed(itemLink))
end

local function VerifyBeltBuckle ()
    return VerifyExtraSocket (INVSLOT_WAIST)
end


------------------------------------------------------------------

local function VerifyItemEnchant (itemLink)
    local _,_,_, ilvl= _G.GetItemInfo(itemLink)
    
    if ilvl==nil or ilvl < sanityCheck_config['EnchMinLvl'] then
        return true
    else
        local itemId, enchantId = itemLink:match("item:(%d+):(%d+)")
        return enchantId~='0'
    end
end

------------------------------------------------------------------

local function VerifyRoleAndSpec ()
    local class = _G.UnitClass('player')
    local spec = _G.GetSpecialization('player')
    local role = _G.UnitGroupRolesAssigned('player')
    
    if tblRoles[class]==nil or spec==nil or _G.UnitGroupRolesAssigned('player')=='NONE' then return true,'' end
    
    if tblRoles[class][spec] == role then
        return true, ''
    else
        local _, spec = _G.GetSpecializationInfo(_G.GetSpecialization())
        return false, 'Are you sure you want to '..tblVerbs[role]..' in '..spec..' spec?'
    end
end

------------------------------------------------------------------

local function VerifyProfessions_Blacksmithing ()
    return VerifyExtraSocket (INVSLOT_HAND) and VerifyExtraSocket (INVSLOT_WRIST)
end

------------------------------------------------------------------

local function VerifyProfessions_Engineering ()
    -- Currently not verifying the belt, because, let's face it, all existing belt tinkers are bullshit :p
    return GetItemSpell(GetInventoryItemLink('player',INVSLOT_HAND)) and GetItemSpell(GetInventoryItemLink('player',INVSLOT_BACK))
end

------------------------------------------------------------------

local tblProfessions = {}
local function InitProfessions ()
    local prof1, prof2 = GetProfessions()

    for i,v in ipairs ({prof1,prof2}) do
        local name, _, rank, maxRank, numSpells, _, skillLine = GetProfessionInfo(v)
        if rank==maxRank then
            tblProfessions[skillLine] = true
        end
    end
    
    -- Enchanters: rings are enchantable
    if tblProfessions[tblProfessionIds.Enchanting] then
        tblSlots_Enchantable[INVSLOT_FINGER1] = true
        tblSlots_Enchantable[INVSLOT_FINGER2] = true
    end
end
------------------------------------------------------------------

local function CheckInventory ()
    if tblWrongStats[_G.UnitClass('player')]==nil then return end
            
    for slot,slotName in _G.pairs(tblSlots) do
        local link = _G.GetInventoryItemLink ('player', slot)
        if link~=nil then
            if not VerifyItemStats(link) or not VerifyItemType(link) then
                _G.table.insert (tblOops[GEAR], 'Bad item: '..slotName)
            end
            if not VerifyItemSockets (link) then
                _G.table.insert (tblOops2[GEAR], 'Missing gems: '..slotName)
            end
            if not VerifySocketAndEnchantStats (link) then
                _G.table.insert (tblOops2[GEAR], 'Bad gems/ enchant: '..slotName)
            end
            
            if tblSlots_Enchantable[slot] and (not VerifyItemEnchant(link)) then
                _G.table.insert (tblOops2[GEAR], 'Missing enchant: '..slotName)
            end
        else
            -- Mention if the item is missing - unless it's an offhand, and there's a 2H equipped
            -- Low level toons can get away with empty slots
            if (slot==INVSLOT_OFFHAND and Has2HanderEquipped()) or (_G.UnitLevel('player')<60) then
                -- All is well
            else
                _G.table.insert (tblOops[GEAR], 'Missing item: '..slotName)
            end
        end
    end
    
    if not VerifyBeltBuckle() then
        _G.table.insert (tblOops2[GEAR], 'Hey, how about a belt buckle and an extra gem?')
    end
    
    if tblProfessions[tblProfessionIds.Blacksmithing] and (not VerifyProfessions_Blacksmithing()) then
        _G.table.insert (tblOops2[GEAR], 'Don\'t forget the extra sockets you can get as a Blacksmith!')
    end

    if tblProfessions[tblProfessionIds.Engineering] and (not VerifyProfessions_Engineering()) then
        _G.table.insert (tblOops2[GEAR], 'Don\'t forget to put some engineering gizmos on your gear!')
    end
end

------------------------------------------------------------------

local function CheckAuras ()
    local class = _G.UnitClass('player')
    local spec = _G.GetSpecialization('player')
    
    -- Personal auras and stuff
    if tblVerifyAuras[class]==nil or spec==nil then return end
    
    local funcs
    local tmp = tblVerifyAuras[class][spec]
    if _G.type(tmp)=='function' then
        funcs = {tmp}
    else
        funcs = tmp
    end
    
    for i,func in _G.ipairs(funcs) do
        local ret, msg = func()
        if not ret then
            _G.table.insert (tblOops[AURA], msg)
        end
    end
    
    -- Raid buffs
    if tblVerifyRaidBuffs[class]==nil then return end
    
    local ret, msg = tblVerifyRaidBuffs[class]()
    if not ret then
        _G.table.insert (tblOops3[AURA], msg)
    end
    
    -- Role
    ret, msg = VerifyRoleAndSpec()
    if not ret then
        _G.table.insert (tblOops[AURA], msg)
    end    
end

------------------------------------------------------------------

local function CreateLightFrame (name, parent)
    local f = CreateFrame ("Frame", name, parent)
    f:SetBackdrop ({
        bgFile = "Interface\\Addons\\SanityCheck\\red.tga",  
        edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
        edgeSize = 8,
    })

    f.lbl = f:CreateFontString(nil, "BACKGROUND", "GameFontNormal")
    f.lbl:SetPoint ("CENTER", f, "CENTER")
    f.lbl:SetSize (100,30)
    f.lbl:SetTextColor (1.0,1.0,1.0)
    f.lbl:SetJustifyH ("CENTER")
    
    f:SetScript("OnEnter", function(self, motion, ...)
        GameTooltip:SetOwner (self, "ANCHOR_TOP")
        self:GenerateTooltip ()
        GameTooltip:Show ()
    end)
    f:SetScript("OnLeave", function(self, motion, ...)
        GameTooltip:Hide ()
    end)

    return f
end

------------------------------------------------------------------


local function BuildGui ()
    -- Primary warning light
    do
        local f = CreateLightFrame ('frmSanityCheck', UIParent)
        f:SetPoint ("CENTER", UIParent, "CENTER", 0, 100)
        f:SetSize (110,40)
        
        f.needGearScan = false
        f.needAuraScan = false
        
        f.elapsed = 0
        f.elapsed2 = 0
        f.itBlinker = iter_cycle ({1.0, 0.0, 0.0, 0.9}, {0.3, 0.0, 0.0, 0.65} )
        f.blinkInterval = 0.55
        f.blink = true
        function f:Light ()
            --self:SetBackdropColor (1.0, 1.0, 1.0, 0.75)
            self.lbl:SetText ("OMG N00b!")
            if self.isOn==false then
                self.blink = true
            end
            self.isOn = true
        end
        function f:Dark ()
            self:SetBackdropColor (0.0, 0.0, 0.0, 0.5)
            self.lbl:SetText ("You're awesome.")
            self.isOn = false
        end

        function f:GenerateTooltip ()
            if self.isOn then
                GameTooltip:AddDoubleLine ('SanityCheck', 'Right-click to stop blinking', 1.0, 1.0, 1.0, 1.0, 1.0, 0.0)
                
                for key,subtbl in _G.pairs(tblOops) do      -- Iterating over the GEAR and AURA sub-tables
                    if #subtbl>0 then
                        for k,v in pairs(subtbl) do
                            GameTooltip:AddLine (v, 1.0, 0.5, 0.5)
                        end
                    end
                end
            else
                GameTooltip:AddLine ('SanityCheck', 1.0, 1.0, 1.0)
                GameTooltip:AddLine ('Your gear and buffs appear to be fine!', 1.0, 0.5, 0.5)
            end
        end

        f:SetMovable (true)
        f:RegisterForDrag ("LeftButton")
        f:SetUserPlaced (true)
        f:SetClampedToScreen (true)
        f:EnableMouse (true)
        f:SetScript("OnMouseDown", function(self, button)
            if button=='LeftButton' then
                self:StartMoving ()
                GameTooltip:Hide()
            elseif button=='RightButton' then
                self.blink = false
            end
        end)
        f:SetScript("OnMouseUp", function(self, button)
            self:StopMovingOrSizing ()
        end)

        f:Dark()
    end

    -- Secondary warning light
    do
        local f = CreateLightFrame ('frmSanityCheck_Secondary', frmSanityCheck)
        f:SetPoint ("LEFT", frmSanityCheck, "RIGHT")
        f:SetSize (20,40)
        
        function f:Light ()
            self:SetBackdropColor (0.9, 0.65, 0.0, 0.85)
            self.lbl:SetText ('!')
            self.isOn = true
        end
        function f:Dark ()
            self:SetBackdropColor (0.0, 0.0, 0.0, 0.5)
            self.lbl:SetText ('')
            self.isOn = false
        end
        f:Dark()
        
        function f:GenerateTooltip ()
            if self.isOn then
                GameTooltip:AddDoubleLine ('SanityCheck', 'Right-click to generate shopping list', 1.0, 1.0, 1.0, 1.0, 1.0, 0.0)
            
                for key,subtbl in _G.pairs(tblOops2) do      -- Iterating over the GEAR and AURA sub-tables
                    for k,v in pairs(subtbl) do
                        GameTooltip:AddLine (v, 1.0, 0.5, 0.5)
                    end
                end
            else
                GameTooltip:AddDoubleLine ('SanityCheck', 'Secondary Warnings', 1.0, 1.0, 1.0, 1.0, 1.0, 0.0)
                GameTooltip:AddLine ('You seem to have all your gems and enchants!', 1.0, 0.5, 0.5)
            end
        end
        
        f:SetScript("OnMouseDown", function(self, button)
            if button=='RightButton' then
                frmShoppingList:Generate()
            end
        end)
       
        frmSanityCheck.frmSec = f
    end

    -- Tertiary warning light
    do
        local f = CreateLightFrame ('frmSanityCheck_Tertiary', frmSanityCheck)
        f:SetPoint ("BOTTOMRIGHT", frmSanityCheck, "BOTTOMLEFT")
        f:SetSize (20,20)
        
        function f:Light ()
            self:SetBackdropColor (0.6, 0.75, 1.0, 0.85)
            self.isOn = true
        end
        function f:Dark ()
            self:SetBackdropColor (0.0, 0.0, 0.0, 0.5)
            self.isOn = false
        end
        f:Dark()
        
        function f:GenerateTooltip ()
            if self.isOn then
                GameTooltip:AddDoubleLine ('SanityCheck', 'Even More Warnings', 1.0, 1.0, 1.0, 1.0, 1.0, 0.0)
            
                for key,subtbl in _G.pairs(tblOops3) do      -- Iterating over the GEAR and AURA sub-tables
                    for k,v in pairs(subtbl) do
                        GameTooltip:AddLine (v, 1.0, 0.5, 0.5)
                    end
                end
            else
                GameTooltip:AddDoubleLine ('SanityCheck', 'Even More Warnings', 1.0, 1.0, 1.0, 1.0, 1.0, 0.0)
                GameTooltip:AddLine ('Nothing to see here, move along now!', 1.0, 0.5, 0.5)
            end
        end
       
        frmSanityCheck.frmTer = f
    end

    --  Configure button
    do
        local f = CreateFrame ("Frame", nil, frmSanityCheck)
        f:SetPoint ("TOPRIGHT", frmSanityCheck, "TOPLEFT")
        f:SetSize (20,20)
        f:SetBackdrop ({
            edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
            edgeSize = 10,
        })
        
        local btn = CreateFrame ("Button", nil, f, "UIPanelButtonTemplate")
        btn:SetAllPoints (f)
        btn:SetFrameStrata("LOW")
        
        btn:SetScript("OnEnter", function(self, motion, ...)
            GameTooltip:SetOwner (self, "ANCHOR_TOP")
            GameTooltip:AddLine ('Configure', 1.0, 1.0, 1.0)
            GameTooltip:Show ()
        end)
        btn:SetScript("OnLeave", function(self, motion, ...)
            GameTooltip:Hide ()
        end)
        btn:SetScript("OnClick", function (self,...)
            frmSanityCheckConfig:Show()
        end)

        frmSanityCheck.btnConfig = btn
    end
end

------------------------------------------------------------------
--  Config frame

local function MakeConfigFrame ()
    local function MakeOptionLabel (frm, txt)
        local lbl = frm:CreateFontString(nil, "BACKGROUND", "GameFontNormal")
        lbl:SetTextColor (1,1,1)
        lbl:SetText (txt)
        lbl:SetJustifyH ("RIGHT")
        
        return lbl
    end
    
    local function MakeOptionTextbox (frm)
        local txt = CreateFrame ("EditBox", nil, frm)
        txt:SetBackdrop ({
            bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background",  
            edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Gold-Border",
            tile = true,
            tileSize = 32,
            edgeSize = 8,
        })
        txt:SetTextColor (1,1,1)
        txt:SetFont ('Fonts\\ARIALN.TTF', 14, "")
        txt:SetAutoFocus (false)
        txt:SetTextInsets (4, 4, 0, 0)
        txt:SetScript ("OnEscapePressed", txt.ClearFocus)
        txt:SetNumeric (true)
        
        return txt
    end
    
    -- Default values
    if g_SanityCheck_config==nil then
        g_SanityCheck_config = {
            EnchMinLvl = 483,
            GemMinLvl = 483,
        }
    end
    
    -- Make frame
    local W, H = 440, 140
    local f = CreateFrame ("Frame", 'frmSanityCheckConfig', frmSanityCheck)
    f:SetPoint ("CENTER", UIParent, "CENTER")
    f:SetSize (W,H)
    f:SetBackdrop ({
        bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background",  
        edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Gold-Border",
        edgeSize = 16,
    })
    
    -- Title
    f.title = f:CreateFontString(nil, "BACKGROUND", "GameFontNormal")
    f.title:SetPoint ("TOPLEFT", f, "TOPLEFT", 0, -4)
    f.title:SetPoint ("TOPRIGHT", f, "TOPRIGHT", 0, -4)
    f.title:SetHeight(20)
    f.title:SetText ('SanityCheck Options')
    
    -- X button
    f.btnX = CreateFrame ("Button", nil, f, "UIPanelCloseButton")
    f.btnX:SetPoint("TOPRIGHT", f, "TOPRIGHT")
    f.btnX:SetSize(24,24)

    -- Options 1: ilvl that requires an enchant    
    f.lblEnchMinLvl = MakeOptionLabel(f, "Don't require enchants on items lower than ilvl:")
    f.lblEnchMinLvl:SetPoint ("TOPRIGHT", f, "TOPRIGHT", -104, -25)
    f.lblEnchMinLvl:SetSize (W-54, 34)
    
    f.txtEnchMinLvl = MakeOptionTextbox (f)
    f.txtEnchMinLvl:SetPoint ("TOPRIGHT", f, "TOPRIGHT", -38, -31)
    f.txtEnchMinLvl:SetSize (46, 22)
    f.txtEnchMinLvl:SetNumber (g_SanityCheck_config['EnchMinLvl'])
    
    -- Options 1: ilvl that requires gems    
    f.lblGemMinLvl = MakeOptionLabel(f, "Don't require gems in items lower than ilvl:")
    f.lblGemMinLvl:SetPoint ("TOPRIGHT", f, "TOPRIGHT", -104, -55)
    f.lblGemMinLvl:SetSize (W-54, 34)
    
    f.txtGemMinLvl = MakeOptionTextbox (f)
    f.txtGemMinLvl:SetPoint ("TOPRIGHT", f, "TOPRIGHT", -38, -61)
    f.txtGemMinLvl:SetSize (46, 22)
    f.txtGemMinLvl:SetNumber (g_SanityCheck_config['GemMinLvl'])
    
    
    -- OK/ Cancel buttons    
    f.btnCancel = CreateFrame ("Button", nil, f, "UIPanelButtonTemplate")
    f.btnCancel:SetPoint("BOTTOMRIGHT", f, "BOTTOMRIGHT", -14, 10)
    f.btnCancel:SetSize(80,24)
    f.btnCancel:SetText ("Cancel")
    f.btnCancel:SetScript ("OnClick", function (self,...) frmSanityCheckConfig:Hide() end)
    
    f.btnOk = CreateFrame ("Button", nil, f, "UIPanelButtonTemplate")
    f.btnOk:SetPoint("RIGHT", f.btnCancel, "LEFT", -7, 0)
    f.btnOk:SetSize(80,24)
    f.btnOk:SetText ("OK")
    f.btnOk:SetScript ("OnClick", function (self,...)
        g_SanityCheck_config['EnchMinLvl'] = frmSanityCheckConfig.txtEnchMinLvl:GetNumber()
        g_SanityCheck_config['GemMinLvl'] = frmSanityCheckConfig.txtGemMinLvl:GetNumber()
        frmSanityCheck.needGearScan = true
        frmSanityCheckConfig:Hide()
    end)
    
    f:Hide()
end

------------------------------------------------------------------
--  Events and such
local function RegisterEvents ()
    local f = frmSanityCheck
    function f:Scan ()
        -- Aura scanning is much lighter than gear scanning, so we always do it
        
        -- Clear tables
        tblOops[AURA] = {}
        tblOops2[AURA] = {}
        tblOops3[AURA] = {}
        if self.needGearScan then
            tblOops[GEAR] = {}
            tblOops2[GEAR] = {}
            tblOops3[GEAR] = {}
            tblShoppingListGems = {}
            tblShoppingListEnch = {}
        end

        -- Now do stuff
        if self.needGearScan and (not inCombat) then
            CheckInventory ()
            self.needGearScan = false
        end
            
        CheckAuras ()
        self.needAuraScan = false
        
        if #tblOops[GEAR] > 0 or #tblOops[AURA] > 0 then
            frmSanityCheck:Light ()
        else
            frmSanityCheck:Dark ()
        end

        if #tblOops2[GEAR] > 0 or #tblOops2[AURA] > 0 then
            frmSanityCheck.frmSec:Light ()
        else
            frmSanityCheck.frmSec:Dark ()
        end

        if #tblOops3[GEAR] > 0 or #tblOops3[AURA] > 0 then
            frmSanityCheck.frmTer:Light ()
        else
            frmSanityCheck.frmTer:Dark ()
        end        
    end
    
    f:RegisterEvent ("UNIT_INVENTORY_CHANGED")
    f:RegisterEvent ("PLAYER_SPECIALIZATION_CHANGED")
    f:RegisterEvent ("UNIT_AURA")
    f:RegisterEvent ("UNIT_PET")
    f:RegisterEvent ("PLAYER_ENTERING_WORLD")
    f:RegisterEvent ("GROUP_ROSTER_UPDATE")   -- To detect role change!
    f:RegisterEvent ("PLAYER_REGEN_ENABLED")
    f:RegisterEvent ("PLAYER_REGEN_DISABLED")
    
    f:SetScript ("OnEvent", function  (self, event, GUID, ...)
        if event=='PLAYER_ENTERING_WORLD' then
            self.needGearScan = true
            self.needAuraScan = true
            inCombat = _G.UnitAffectingCombat('player')
        elseif event=='PLAYER_REGEN_DISABLED' then
            inCombat = true
        elseif event=='PLAYER_REGEN_ENABLED' then
            inCombat = false
        elseif event=='UNIT_INVENTORY_CHANGED' and GUID=='player' then
            self.needGearScan = true
        elseif (event=='UNIT_PET' or event=='UNIT_AURA') and GUID=='player' then
            self.needAuraScan = true
        elseif event=='PLAYER_SPECIALIZATION_CHANGED' then
            self.needGearScan = true
            self.needAuraScan = true
        end        
    end)
    f:SetScript ("OnUpdate", function (self, elapsed, ...)
        self.elapsed2 = self.elapsed2 + elapsed
        
        --if self.elapsed2 > THROTTLE and (self.needGearScan or self.needAuraScan) then
        if ((not self.needGearScan) and self.needAuraScan and self.elapsed2 > THROTTLE)  or  (self.needGearScan and self.elapsed2 > THROTTLE * 3) then
            self:Scan ()
            self.elapsed2 = 0
        end

        if self.isOn then
            self.elapsed = self.elapsed + elapsed
            
            if self.elapsed > self.blinkInterval then
                self.elapsed = 0
                if self.blink then
                    self:SetBackdropColor (_G.unpack(self.itBlinker:next()))
                else
                    self:SetBackdropColor (_G.unpack(self.itBlinker.values[1]))
                end
            end
        end        
    end)
end


------------------------------------------------------------------

function CreateShoppingList ()
    -- A point!
    -- Because the frame gets anchored to the fontstring, and fontstrings apparently can't be moved (?)
    local point = CreateFrame ("Frame", 'frmShoppingList_point', UIParent)
    point:SetSize (1,1)
    point:SetPoint ("CENTER", UIParent, "CENTER")
    
    local f = CreateFrame ("Frame", 'frmShoppingList', UIParent)
    f:SetBackdrop ({
        bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background",
        edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Gold-Border",
        edgeSize = 8,
    })

    f.lbl = f:CreateFontString(nil, "BACKGROUND", "GameFontNormal")
    f.lbl:SetPoint ("CENTER", point, "CENTER")
    f.lbl:SetTextColor (1.0,1.0,1.0)
    f.lbl:SetJustifyH ("LEFT")
    
    f:SetPoint ("TOPLEFT", f.lbl, "TOPLEFT", -12, 12)
    f:SetPoint ("BOTTOMRIGHT", f.lbl, "BOTTOMRIGHT", 12, -12)
    f:Hide ()
    
    -- Make movable
    point:SetMovable (true)
    point:RegisterForDrag ("LeftButton")
    point:SetUserPlaced (true)
    point:SetClampedToScreen (true)
    point:EnableMouse (true)
    f:SetScript("OnMouseDown", function(self, button)
        if button=='LeftButton' then
            frmShoppingList_point:StartMoving ()
        elseif button=='RightButton' then
            frmShoppingList_point:StopMovingOrSizing ()
            self:Hide()
        end
    end)
    f:SetScript("OnMouseUp", function(self, button)
        frmShoppingList_point:StopMovingOrSizing ()
    end)

    function f:Generate ()
        for slot,slotName in _G.pairs(tblSlots) do
            local link = _G.GetInventoryItemLink ('player', slot)
            if link~=nil then
                if tblSlots_Enchantable[slot] and (not VerifyItemEnchant(link)) then
                    tblShoppingListEnch[slotName] = 1
                end
                
                ShoppingList_CountSockets (link)
            end
        end    
    
        local txt = 'Empty sockets:\n'
        for k,v in pairs(tblShoppingListGems) do
            txt = txt..v..'x  '..k..'\n'
        end
        txt = txt..'\nMissing enchants:\n'
        for k,v in pairs(tblShoppingListEnch) do
            txt = txt..k..'\n'
        end
        self.lbl:SetText (txt)
        
        self:Show()
    end
end

------------------------------------------------------------------
------------------------------------------------------------------

BuildGui ()
CreateShoppingList ()

local starter = CreateFrame ("Frame")
starter:RegisterEvent ("PLAYER_ENTERING_WORLD")
starter:SetScript ("OnEvent", function  (self, e, ...)
    MakeConfigFrame ()
    RegisterEvents ()
    InitProfessions ()
    frmSanityCheck:GetScript ('OnEvent') (frmSanityCheck, 'PLAYER_ENTERING_WORLD')
    sanityCheck_config = g_SanityCheck_config
end)